; Functions to be called from C to access 386 protected mode
; GPL 2 software by Eric Auer <eric@coli.uni-sb.de> 2001

; Compile: nasm -fobj pmlib.asm
; No assumptions made: CS has to be readable, segs have to be
; use16, offsets are arbitrary. Some funcs assume 40h to be
; a flat 4gb data segment, however.

; for allocation, we could either hook the int 15 memsize stuff
; (and return less) or alloc our memory from XMS, which seems to
; be common among EMM386-like software. So I include XMS alloc
; and free stuff here :-)


%define simplistic_prot 1
; ^- define to do a 2nd goprot attempt if int 15.89 fails:
;    simplistic because I did not implement A20 or 8259 stuff there


SEGMENT _TEXT ALIGN=4 CLASS=CODE
GLOBAL _is386, _goprot, _ramsize, _copyflat, _setpmstuff, _getxms, _mkidt
GLOBAL _fallbacktoreal

; --------------------------------------------------------------------

%imacro AXtoFtoAX 0	; see if we can load ax to the flags
	push ax
	popf
	pushf
	pop ax
%endmacro

; *** int is386(int needpm)
;
; Returns 1 if a 386 is found, but if needpm!=1, only if
; protected mode is not already on. Else returns 0.
;
_is386:
	push bp
	mov bp,sp
	pushf
	cli
	        xor ax,ax
		AXtoFtoAX	; try to clear all bits
	and ax,0f000h
	cmp ax,0f000h
	jz short _is386n	; 4 msb stuck to 1: 808x or 80186
		mov ax,0f000h
		AXtoFtoAX	; try to set the high nibble bits
	test ax,0f000h
	jz short _is386n	; 4 msb stuck to 0: 80286
		mov ax,[bp+4]	; needpm
	or ax,ax
	jz short _is386y	; if needpm==0, we return success
		smsw ax		; MSW is the low half of CR0
	test ax,1
	jnz _is386n		; lsb is 1: protected mode already on
_is386y:
	mov ax,1
	jmp short _is386x
_is386n:			; return failure
	xor ax,ax		
_is386x:
	popf
	pop bp
	ret			; return result in AX

; --------------------------------------------------------------------

; modes: PxxYzzzz, where P:present xx:DPL Y:1app/0sys zzzz:type,
; where for sys, types are: 0nil 1tss16 2ldt 3tss16busy
; 4callgate16 5taskgate 6intgate16 7trapgate16 (10--- 13---)
; 8nil 9tss32 11tss32busy 12callgate32 14intgate32 15trapgate32
; and for app, types are: 0..7:ds: down,wr,used 8..f:cs: conf,rd,used

%imacro DOStoDESC 0	; make a DS descriptor from seg AX at ds:di
	mov word [di],-1	; limit
	push eax
	movzx eax,ax
	shl eax,4		; para -> flat
	mov dword [di+2],eax	; base (24bit)
	mov word [di+5],93h	; low: mode hi: GD0xLLLL, where
	shr eax,24		; G=shl12 D=use32 (x=free) L=limitmsb
	mov byte [di+7],al	; msb of base ^- hi is 00 for 286 segs
	pop eax			; (seg is DPL 0, up, writeable, used)
%endmacro

; *** int goprot(void near *gdt, void near *idt, unsigned int irqpos)
;
; converts the current segments to GDT descriptors and remaps
; irq 0..7 to "from irqpos lo" and irq 8..f to "from irqpos hi".
; this uses the BIOS function int 15.89, which does most of the
; work for us (8259 irq reprogramming, A20 enabling, seg setup).
; returns ah=0 on success, ah=ff on error (a20, irq, ...).
; we also provide a flat DS (40h) and a flat CS (48h), so 
; the first free GDT slot will be at offset 50h!

_goprot:
	push bp
	mov bp,sp
	push bp		; destroyed by BIOS
	push bx
	push si		; ES set to DS as ES:SI -> GDT...
		mov ax,ds
		mov es,ax
	mov si,[bp+4]	; * gdt
	mov bx,[bp+8]	; irqpos
	and bx,0f8f8h	; sanitize
	cmp bl,bh
	jnz _gopbxok
	mov bx,07008h	; if bx was insane, set to DOS defaults
_gopbxok:
	push eax
	push di
	xor eax,eax
	mov [si],eax	; NUL1
	mov [si+4],eax	; NUL2
	;
	mov ax,[bp+4]	; * gdt
	shr ax,4
	mov di,ds
	add ax,di	; dos seg for GDT
	lea di,[si+8]	; start here
	DOStoDESC	; alias dseg for GDT
	mov ax,[bp+4]	; * gdt
	and ax,15
	or [di+2],ax	; allow misalignment
	mov word [di],1ffh	; limit: 0..(64*descriptor size)
	;
	add di,8
	mov ax,[bp+6]	; * idt
	shr ax,4	; <-
		push di
	mov di,ds
	add ax,di	; dos seg for IDT
		pop di
	DOStoDESC	; alias dseg for IDT
	mov ax,[bp+6]	; * idt
	and ax,15
	or [di+2],ax	; allow misalignment
	mov word [di],7ffh	; limit: 0..(256*descriptor size)
	;
	add di,8
	mov ax,ds
	DOStoDESC	; DS, protected
	add di,8
	mov ax,es
	DOStoDESC	; ES, protected
	add di,8
	mov ax,ss
	DOStoDESC	; SS, protected
	add di,8
	mov ax,cs
	DOStoDESC	; CS, protected
	mov byte [di+5],9bh	; make a CS, readable, nonconf, used
	add di,8
	mov ax,0f000h
	DOStoDESC	; BIOS CS (done by BIOS anyhow)
	mov byte [di+5],9bh	; make a CS...
	add di,8
	mov eax,0000ffffh	; limit -1, base 0
	mov [di],eax	; flat DS1
	mov [di+8],eax	; flat CS1
	mov eax,008f9300h	; so base 0, limit (0fffffh<<12)+0fffh
	mov [di+4],eax	; flat DS2	; segs are use16 (for Turbo C)
	mov ah,9bh		; make a DS...
	mov [di+12],eax	; flat CS2
	pop di
	pop eax
	;
	push si
	mov ah,89h
	int 15h		; go for it!
	pop si
	jnc near _gopok	; on error: CF set, ah=ff

;-------

%ifdef simplistic_prot
		smsw ax		; protected mode already on?
	test ax,1
	jnz near _gopfail	; don't be silly in that case!

;-------
	
	call _gop_where
_gop_where:
	pop bp			; assume a use16 CS, as usual!
	sub bp, _gop_where	; adjust...
	add bp, _gopok		; ...jump destination in BP
	push ebx	; save...
	mov ebx,[si+10h+2]	; the IDT alias desriptor -> base
			; *** 8/2003: must be 10h, not 10... ***
	and ebx,0ffffffh	; lazy: IDT must be < 16 MBy
	mov eax,[si+8+2]	; flat pointer to GDT
	and eax,0ffffffh	; lazy: GDT must be < 16 MBy
	push eax	; hi: base: flat pointer to GDT
	push word 1ffh	; lo: size: allow 64 descriptors
	;
	movzx esp,sp	; KLUDGE WARNING !
	mov al,80h
	out 70h,al	; disable NMI
	cli		; disable IRQ
	;
		lgdt [esp+0]	; initialize GDT
		mov eax,cr0	; machine status...
	or al,1			; enable protected mode, PE
		mov cr0,eax	; START protected mode
	jmp short _goprflush	; flush queue
_goprflush:			; gdt: nul, gdtalias, idtalias, ds,
	;			; es, ss, cs, bioscs, flatds, flatcs
	mov ax,18h
	mov ds,ax
	mov ax,20h
	mov es,ax
	mov ax,28h
	mov ss,ax		; has to be an alias of our old SS
	mov ax,40h
	mov fs,ax		; flat DATASEG
	mov gs,ax		; flat DATASEG
	movzx esp,sp		; remind the stack of the change...
	;
	push ebx		; IDT position (read from GDT above)
	push word 7ffh		; 256 vectors/descriptors
		lidt [esp+0]
	pop ax			; -
	pop eax			; -
		lgdt [esp+0]	; reload gdt
	pop ax			; -
	pop eax			; -
	pop ebx		; ...restore
	;
	xor eax,eax		; return status: success
	push word 30h		; new CS
	push bp			; prepared to point to _gopok
	retf			; activate new CS and jump to _gopok!
%endif

;-------

_gopfail:
	mov ax,-1
_gopok:
	pop si
	pop bx
	pop bp
	pop bp
	ret

; --------------------------------------------------------------------

; *** unsigned int ramsize(dword near * under16m, 
; ***	dword near * above16m)
; return amount of RAM beyond 0x100000 (1mb)
; we do not use int 15.c7 (return sizes of local/sys/cacheable/nonsys
; memory, separate for 1..16m and 16m..4g, bla), nor the phoenix func
; 15.8a->dxax, nor ami func 15.da88->clbx, BUT 15.88, and 15.e801;
; the advanced 15.e820 memory map is not yet supported.
; under16m and above16m returns the kbytes installed <16mb and >16mb.

_ramsize:
	push bp
	mov bp,sp
	mov ax,8855h	; request memory size (will tell max. 15mb)
	int 15h		; (max 15mb because 1st mb is excluded)
	jc _ramsize0
	cmp al,55h
	jnz _ramsizex	; bonus sanity check
_ramsize0:
	xor ax,ax
_ramsizex:
	pusha
	push eax
	mov di,[bp+4]	; * under16m
	xor eax,eax
	mov [di],eax	; start with 0
	mov di,[bp+6]	; * above16m
	mov [di],eax	; start with 0
	mov ax,0e801h	; request memory size big
	int 15h
	jc _ramszx2
	cmp ax,2000h	; 8m (if < 3c00/15m: hole exists)
	jnb _ramsz_axbx
_ramsz_cxdx:		; try alternative values
	mov ax,cx
	mov bx,dx
_ramsz_axbx:
	mov di,[bp+4]	; * under16m
	movzx eax,ax
	mov [di],eax	; under16m
	mov di,[bp+4]	; * above16m
	movzx eax,bx
	shl eax,6	; unit was 64k, now it is 1k
	mov [di],eax	; above16m
_ramszx2:
	pop eax
	popa
	pop bp
	ret

; --------------------------------------------------------------------

; *** int copyflat(dword src, dword dst, dword size)
; copies size bytes around in flat memory space, assuming
; 40h to be a flat DS - and if protected mode is off, we
; use the BIOS (other software would decide the inverse!)
; returns 0 on success

_copyflat:
	push bp
	mov bp,sp
	smsw ax
	test ax,1
	jz _cpyreal
	push ds
	push es
	push esi
	push edi
	push ecx
	mov esi,[bp+4]	; src
	mov edi,[bp+8]	; dst
	mov ecx,[bp+12]	; size
	mov ax,40h
	mov ds,ax
	mov es,ax
	test cl,1
	jz _cpyeven1
	mov al,[ds:esi]
	mov [ds:edi],al
	inc esi
	inc edi
_cpyeven1:
	test cl,2
	jz _cpyeven2
	mov ax,[ds:esi]
	mov [ds:edi],ax
	inc esi
	inc esi
	inc edi
	inc edi
_cpyeven2:
	shr ecx,2	; bytes -> dwords
	jecxz _cpydone
	pushf
	cld
	a32 rep movsd 	; main copy loop !
			; a32 -> use edi/esi
			; es for di is fixed, ds for si not...!
			; using edi/esi triggers using ecx as well!
	popf
_cpydone:
	xor ax,ax	; success (why not :-))
	pop ecx
	pop edi
	pop esi
	pop es
	pop ds
_cpyflatx:
	pop bp
	ret

;-------

_cpyreal:		; so this sucks, but whatever...
	push es
	push si
	push ecx
	mov ax,ss
	mov es,ax	; for our GDT will be on stack
	mov ax,-1
	test byte [bp+12],1	; size odd is forbidden
	jnz _cpyrdone
	cmp word [bp+12+2],1	; sizeHI
	ja _cpyrdone
	jb _cpyrok
	cmp word [bp+12],0	; size exactly 64k is ok as well
	jnz _cpyrdone
_cpyrok:
	mov ecx,[bp+12]
	shr ecx,1	; count in words (BIOS: <= 32k)
	sub sp,48	; for GDT
	mov si,sp	; pointer to GDT
	push eax
	xor eax,eax	; <-
	mov [si],eax	; NUL
	mov [si+4],eax	; NUL
	mov [si+8],eax	; -
	mov [si+12],eax	; -
	mov [si+32],eax	; bioscs
	mov [si+36],eax	; bioscs
	mov [si+40],eax	; biosss
	mov [si+44],eax	; biosss
	dec eax
	mov [si+16],ax	; src limit
	mov [si+24],ax	; dst limit
	mov eax,[bp+4]	; src
	mov [si+18],eax	; src (24lsb)
	shr eax,24
	mov [si+23],al	; src (8msb)
	mov eax,[bp+8]	; dst
	mov [si+26],eax	; dst (24lsb)
	shr eax,24
	mov [si+31],al	; dst (8msb)
	mov ax,8f93h	; (f: 4msb limit -> fffff -> gran -> ffffffff)
	mov [si+21],ax	; src: 8f - granular16/f and 93 - a DS
	mov [si+29],ax	; dst: 8f - granular16/f and 93 - a DS
	pop eax
	mov ah,87h
	int 15h		; BIOS: copy extended memory
	add sp,48	; remove GDT from stack
			; returned status is in AX
_cpyrdone:
	pop ecx
	pop si
	pop es
	jmp _cpyflatx	; return

; --------------------------------------------------------------------

; *** int setpmstuff(void near * gdtp, void near * idtp,
; ***		int ldt, unsigned long cr3,
; ***		int tss, int oldtss)
; to be called from protected mode!
; reload gdt and idt using the given DS pointers to
; { WORD limit, DWORD flatpos } STRUCTURES (!),
; load the given ldt if not 0, activate the given tss if not 0,
; and activate the given page table (cr3) if not 0.
; Give a cr3 of -1 to turn off paging.
; If we have an oldtss, it will be activated by LTR and marked active
; before the new tss is acativated - so the oldtss must be inactive
; on calling this function!
; Obviously, if TSS is not 0, LDT and CR3 gets overridden
; by the values from TSS - at least for a 32bit TSS!
; Returns near pointer to _gotnotss if a TSS of -1 is given!
; the oldtss descriptor must, the tss decriptor must not be busy.

_setpmstuff:
	push bp
	mov bp,sp
	push eax
	mov ax,[bp+4]	; gdt
	or ax,ax
	jz _gotnogdt
_gotgdt:
	movzx eax,ax
		lgdt [eax]	; LOAD GDTR from struc in ram
_gotnogdt:
	mov ax,[bp+6]	; idt
	or ax,ax
	jz _gotnoidt
_gotidt:
	movzx eax,ax
		lidt [eax]	; LOAD IDTR from struc in ram
_gotnoidt:
	mov ax,[bp+8]	; ldt
	or ax,ax
	jz _gotnoldt
_gotldt:
		lldt ax		; LOAD LDT (a selector)
_gotnoldt:
	mov eax,[bp+10]	; cr3
	or eax,eax
	jz _gotnocr3
	cmp eax,-1
	jnz _gotcr3
_gotcr3off:
		mov eax,cr0	; cr0 ->
	and eax,7fffffffh
		mov eax,cr0	; -> turn off paging
	xor eax,eax
		mov cr3,eax	; flush paging system
	jmp short _gotnocr3
_gotcr3:
		mov eax,cr0	; cr0 ->
	or eax,80000001h
		mov cr0,eax	; -> enable paging and protected!
		mov cr3,eax	; LOAD CR3 (flat ptr to paging tables)
_gotnocr3:

	mov ax,[bp+14]	; tss
	or ax,ax
	jz _gotnotss
	cmp ax,-1
	jz _gottofindme
_gottss:
	push ax		; will be CS below
	mov ax,[bp+16]	; oldtss
	or ax,ax
	jz _gotnooldtss
_gotoldtss:
	push bx
	mov bx,ax
		str ax		; ASK OLD TSS
			; not privileged but not allowed in v86
	cmp ax,bx
	mov ax,bx
	pop bx
	jz _gotnooldtss	; we may not activate a TSS twice!
		ltr ax		; SET OLD TSS (TR) - no task switch
_gotnooldtss:
	xor ax,ax	; IP is irrelevant
	push ax
		retf		; LOAD NEW TSS into CS
			; This will make the full current status
			; go into the OLD TSS !
			; (If the OLD TSS gets active again, we
			; continue with _gottofindme)
			; The NEW TSS should probably do CLTS to
			; indicate that we did not "dirty the FPU" ?!
_gottofindme:
	pop eax
	call _gotorientation
_gotorientation:
	pop ax		; assume a USE16 CS (as throughout this file!)
	add ax, _gotnotss - _gotorientation	; reloc'able :-)
	push eax
_gotnotss:
	pop eax
	pop bp
	ret

; to load gdtr or idtr, we need in RAM a structure of word limit,
; dword flatoffset (limit: (8*numdesc)-1).

; --------------------------------------------------------------------

%imacro GOtoXMS 0	; call XMS driver ... if ax is returned
			; as 0, jump to _getxmsfail
	call far [bp-4]	; locally stored pointer!
	or ax,ax
	jz near _getxmsfail
%endmacro

; *** int getxms(unsigned long int size, unsigned long int near * pos,
; ***		unsigned long int near * remaining)
; allocate size bytes of XMS and lock them in RAM.
; returns handle, and sets flat position of that XMS in pos.
; if size is < 0, it is taken to be -(a handle number to free).
; returns 0 on failure, and more details in pos/lo (!).
; remaining tells us how much XMS is left... (in kbytes).

_getxms:
	push bp
	mov bp,sp
	push dword 12345678h	; local var
	push bx
	push edx
	push eax
	mov ax,4300h
	mov bl,43h	; prepare an error code
	int 2fh		; XMS installation check
	cmp al,80h
	jnz near _getxmsfail
	push es
	mov ax,4310h
	int 2fh		; XMS get driver entry point ESBX
	mov [bp-4],bx	; local...
	mov [bp-2],es	; ...copy of that pointer
	pop es
	;
	mov eax,[bp+4]	; size
	or eax,eax
	js _getxmsfree	; FREE some memory
_getxmsalloc:		; ALLOC some memory
	add eax,1023
	shr eax,10	; convert to kbytes
	test eax,0ffff0000h	; need to use XMS 3, XXMS ?
	jz _getsmallxms

_getbigxms:
	; we could use func 0, get version ax / subvers bx / hma dx
	; here: the big-{alloc, query, handleinfo, realloc}
	; and optional realloc umb func are only avail there
	mov edx,eax	; size in kbytes
	mov ah,89h	; allocate memory
	GOtoXMS		; handle: DX
	jmp short _getxmslock

_getsmallxms:
	mov dx,ax	; size in kbytes
	mov ah,9	; alloc block
	GOtoXMS		; handle: DX

_getxmslock:
	mov ah,0ch	; lock block (dx)
	mov bx,[bp+8]	; ptr to pos
	mov [bx+2],dx	; save handle!
	GOtoXMS
	mov ax,bx	; (pos lo)
	mov bx,[bp+8]	; ptr to pos
	mov [bx],ax	; flat pos lo
	xchg [bx+2],dx	; flat pos hi, restore handle!
	jmp short _getxmsx1

_getxmsfree:
	neg eax		; get handle number
	mov dx,ax
	mov bx,[bp+8]	; ptr to pos
	mov [bx],dx	; save handle!
	mov ah,0dh	; unlock block (dx)
	GOtoXMS
	mov bx,[bp+8]	; ptr to pos
	mov dx,[bx]	; restore handle!
	mov ah,0ah	; free block (dx)
	GOtoXMS

_getxmsx1:
	mov al,0	; no error
	jmp _getxmsremaining

_getxmsfail:
	xor dx,dx	; handle zero for error
	movzx eax,bl	; xms error code
	mov bx,[bp+8]	; ptr to pos
	mov [bx],eax	; error code

;-------

_getxmsremaining:
	mov bx,[bp+10]	; ptr to remaining
	cmp al,43h	; no xms found?
	jz _getxmsnorem
	xor eax,eax
	mov [bx],eax	; assume 0 rem first
	;
	mov ah,8
	mov bl,0	; (from intlist)
	push dx
	call far [bp-4]	; query remaining
	pop dx
	or ax,ax
	jz _getxmsnorem	; assume 0 if error
	movzx eax,ax
	mov bx,[bp+10]	; ptr to remaining
	mov [bx],eax	; largest block, in KBy
	;
	mov ah,88h	; big version, same call
	push edx
	push ecx
	call far [bp-4]	; query big
	pop ecx		; not asking for mem limit
	pop edx		; not asking for total kbytes
	or bl,bl
	jnz _getxmswhatrem	; skip if error 
	mov bx,[bp+10]	; ptr to remaining
	mov [bx],eax	; largest block, in kby

_getxmswhatrem:
	pop eax
	mov ax,dx	; return the handle number
	pop edx
	pop bx
	add sp,4	; remove local var
	pop bp
	ret

_getxmsnorem:
	xor eax,eax
	mov [bx],eax
	jmp short _getxmswhatrem

; --------------------------------------------------------------------

%imacro XmultiHANDwo 2	; 1st: starting number 2nd: number of handlers
%assign i %1		; creates handlers for ints w/o error code.
%rep %2			; repeat
	push dword i<<16	; pseudo error code: int num
	jmp near _thehandler	; just jump to main handler
align 16			; make things easier to count
%assign i i+1		; on to the next handler
%endrep
%endmacro

%imacro mkIDTentry 0	; cs:eax -> int gate at [es:bx], bx+=8
	mov [es:bx],ax			; EIPlo
	mov [es:bx+2],cs		; CS - do again if CS changes!
					; must be CS of _thehandler...
	mov word [es:bx+4],08e00h	; P, PL 0, int gate (8f: trap)
	rol eax,16			; must be use32 to handle v86
					; ints -> 8e/8f, not 86/87!
	mov [es:bx+6],ax		; EIPhi
	ror eax,16
	add bx,8
%endmacro

; *** int mkidt(void far * chandler, void far * idtpos, int handlerds)
; Creates and populates the IDT.
; Modifies itself -> call with DS being an alias for CS !
; The chandler gets called as ---> void handler(hstruct far * hregs);
; and may/should (for faults), but needs not (for traps == {int nn,
; into, int3, some int 1} !) change hregs.
; hregs looks like this:     (*=only valid if cs has different PL)
; DW handlerip, DW handlercs,
; DD edi, esi, ebp, espPM,
; DD ebx, edx, ecx, eax,
; DW es-, ds-, fs-, gs-,	(-=from PM, boring if eflags.vm set)
; DW errorcode, DW intnum,
; DD eip, cs, eflags, esp*, ss*,
; DD es+, ds+, fs+, gs+         (+=only valid if eflags.vm is set)

_mkidt:				; idt may contain int/trap/task gates
	push bp
	mov bp,sp
	push eax
	push bx
	push cx

	call _mkwhereami
_mkwhereami:
	pop ax			; use16 as usual

	push ax
	add ax, _pokethehandler - _mkwhereami
	mov bx,ax
	mov eax,[bp+4]		; chandler
	mov [bx-4],eax		; SELFMODIFYING CODE, warning!
	pop ax
	push ax
	add ax, _poketheseg - _mkwhereami
	mov bx,ax
	mov ax,[bp+12]		; handlerds
			; *** 8/2003: must be +12, not +10 ***
	mov [bx-2],ax		; SELFMODIFYING CODE, warning!
	pop ax
	les bx,[bp+8]		; idtpos
	add ax, _int0 - _mkwhereami
	movzx eax,ax		; EIP of handler
	mov cx,256		; size...

_mkidtloop:
	mkIDTentry
	add eax,10h		; <- just assume 16bytes per handler !!!
	dec cx
	jnz _mkidtloop

	pop cx
	pop bx
	pop eax
	pop bp
	ret

; ------

_thehandler:		; the big generic int handler!
	push gs
	push fs
	push ds
	push es
	pushad
	push dword 12345678h	; the handler (far pointer)
_pokethehandler:
	mov bp,sp	; points to what-we-pushed-last
	push ss		; segment of relevant data
	push bp		; offset of relevant data
	mov ax,1234h		; the handler DS
_poketheseg:
	mov ds,ax
	mov es,ax
	mov fs,ax
	mov gs,ax
	call far [bp]
	add sp,8	; remove args (c calling convention) and
			; the pointer to the handler
	popad
	pop es
	pop ds
	pop fs
	pop gs
	add sp,4	; remove error code
	iretd		; IRETD because the gate is/must be use32
			; (but our CS does not need to be use32)

	jmp short _thehandler	; for int-in-an-own-task, we loop!

; ------

align 16	; do also START aligned! (needed for calc)

_int0:	XmultiHANDwo 0,8	; 8 normal handlers 0..7

_int8:
	movzx esp,sp		; UGLY CODE warning!
	mov word [esp+2],8	; add pseudo error code: int num
	jmp near _thehandler	; (pseudo error code ends up in
align 16				; high word of real error code)
	
	XmultiHANDwo 9,1

_int10:
	movzx esp,sp		; UGLY CODE warning!
	mov word [esp+2],10	; add pseudo error code: int num
	jmp _thehandler		; (pseudo error code ends up in
align 16			; high word of real error code)

_int11:
	movzx esp,sp		; UGLY CODE warning!
	mov word [esp+2],11	; add pseudo error code: int num
	jmp _thehandler		; (pseudo error code ends up in
align 16			; high word of real error code)

_int12:
	movzx esp,sp		; UGLY CODE warning!
	mov word [esp+2],12	; add pseudo error code: int num
	jmp _thehandler		; (pseudo error code ends up in
align 16			; high word of real error code)

_int13:
	movzx esp,sp		; UGLY CODE warning!
	mov word [esp+2],13	; add pseudo error code: int num
	jmp _thehandler		; (pseudo error code ends up in
align 16			; high word of real error code)

_int14:
	movzx esp,sp		; UGLY CODE warning!
	mov word [esp+2],13	; add pseudo error code: int num
	jmp _thehandler		; (pseudo error code ends up in
align 16			; high word of real error code)

	XmultiHANDwo 15,240	; 240 remaining handlers 15..255

; --------------------------------------------------------------------

%imacro GDTtoDOS 0 ; GDT on fs:eax, selector bx -> dos seg (e)bx
	push eax
	movzx ebx,bx
	and bl,0f8h	; we COULD do something to handle the LDT,
	mov eax,[fs:eax+ebx+2]	; but we (*ASSUME*) no LDT usage!
	and eax,0ffffffh	; we COULD check the other 8bits
	shr eax,4	; on their strange place, but we cannot
			; handle > 1 MB anyway! We (*ASSUME*) the
	mov ebx,eax	; seg to start at 000xxxx0h, warning!
	pop eax
%endmacro

; *** int fallbacktoreal(int killflat);
; brings us back to real mode, making some sanity assumptions
; the caller has to meet (not checked). If killflat is != 0,
; FS and GS will be reset to 0.

_fallbacktoreal:
	push bp
	mov bp,sp
	mov ax,[bp+4]	; killflat
	or ax,ax
	jz _fb2r_flat
	xor ax,ax
	mov fs,ax	; kill those...
	mov gs,ax	; ...flat segments
_fb2r_flat:
	mov eax,cr0	; system status
	and eax,7fffffffh
	mov cr0,eax	; turn PG off (no more paging)
	xor eax,eax
	mov cr3,eax	; flush paging system
	; next, we convert all segs to real mode compatible
	; segs: (*ASSUMPTION*) to be taken care for by the caller!
	movzx esp,sp	; make stack 16bit like
	push eax	; hi: base
	push ax		; lo: limit
		sgdt [esp+0]	; snoop on GDT
	pop ax		; -
	pop eax		; GDT linear address
	;
	pushad
	push fs
	push word 40h	; last use of our flat DS (*ASSUMPTION*)
	pop fs
	mov bx,ds	; convert DS
	GDTtoDOS
	mov dx,bx	; remember DS in EDXhi
	mov bx,es	; convert ES
	GDTtoDOS
	shl edx,16
	mov dx,bx	; remember ES in DX
	mov bx,ss	; convert SS
	GDTtoDOS
	mov cx,bx	; remember SS in ECXhi
	mov bx,cs	; convert CS
	GDTtoDOS
	shl ecx,16
	mov cx,bx	; remember CS in CX
	pop fs		; bye bye FS...
	;
	call _fb2r_where
_fb2r_where:
	pop bx		; assume use16, as usual
	sub bx,_fb2r_where
	add bx,_fb2r_jump
	mov al,80h
	out 70h,al		; no NMI here
	pushf
	cli
	push cx	; CSnew
	push bx	; IPnew
		mov eax,cr0	; machine status
	and eax,7ffffffeh	; disable PE and PG (unpaged, real!)
		mov cr0,eax	; we are feeling very real now!
	retf			; flush that CS and the queue!
				; (stack works: descriptor cache)
_fb2r_jump:
	push dword 0	; IDT real mode base
	push dword 3ffh	; IDT real mode limit (4*256, not 8*...!)
		lidt [esp+0]	; activate real mode IDT
	add sp,6	; <- remove idt-where-struct from stack
	mov es,dx		; real mode ES from DX
	shr edx,16
	mov ds,dx		; real mode DS from EDXhi
	shr ecx,16
	mov ss,cx		; real mode SS from ECXhi
	popf
	mov al,0
	out 70h,al		; allow NMI again
	popad
	;
	pop bp
	ret			; according to theory, we are back now!


